#!/bin/bash

export LANG=POSIX
export LC_ALL=POSIX

if [ ! -e /etc/psa/psa.conf ] ; then
  echo "FATAL: Plesk is not installed on this server." >&2
  exit 1
fi

PRODUCT_ROOT_D=`grep ^PRODUCT_ROOT_D /etc/psa/psa.conf | awk '{print $2}'`
HTTPD_VHOSTS_D=`grep ^HTTPD_VHOSTS_D /etc/psa/psa.conf | awk '{print $2}'`
CHROOT_ROOT_D="$HTTPD_VHOSTS_D/chroot"

usage() {
    cat <<HELP
Manage chrooted template in $CHROOT_ROOT_D and apply it to domains.

Usage:
  $0 --install
  $0 --update
    Create default chrooted environment or update existing one
    using actual libraries from the system.

  $0 --add <path>
    Add program to chrooted environment. All linked libraries
    will be copied as well. Program will always be put in /bin
    directory of chrooted environment.

  $0 --devices
    Add additional devices to chrooted environment template. List
    of devices added includes: /dev/tty, /dev/urandom, /dev/ptmx 
    and /dev/zero. Useful for some commands like 'rsync', 'ssh' etc.

  $0 --remove <domain,...>
  $0 --remove all
    Remove chrooted environment from specific domains (comma-separated
    list) or all domains with enabled chrooted shell if no domains are
    given. Useful if full re-initialization of chrooted shell environment
    on domain is needed.

  $0 --apply <domain,...>
  $0 --apply all
    Apply new chrooted template to specific domains (comma-separated
    list) or all domains with enabled chrooted shell if no domains 
    are given. This operation is necessary to apply changes done by 
    '--install', '--add' and '--devices' commands to specific or all
    domains.

HELP
}

# Do mysql query on Plesk database
q() {
  local query="$1"

  mysql -Ns -uadmin -p$(cat /etc/psa/.psa.shadow) -Dpsa -e"$query"
}

# get plesk version
plesk_version() {
  awk -F. '{print $1$2}' $PRODUCT_ROOT_D/version
}

# Get all domains
all_domains() {
  if [ `plesk_version` -ge 100 ] ; then
        q "SELECT d.name FROM domains d, hosting h, sys_users s WHERE d.id = h.dom_id AND h.sys_user_id = s.id AND s.shell = '$PRODUCT_ROOT_D/bin/chrootsh' AND d.webspace_id = 0"
  else
        q "SELECT d.name FROM domains d, hosting h, sys_users s WHERE d.id = h.dom_id AND h.sys_user_id = s.id AND s.shell = '$PRODUCT_ROOT_D/bin/chrootsh'"
  fi
}

# Description: Copy all linked libraries of binary into chrooted environment
# Parameters:
# $1: binary itself
# $2: path to chrooted /lib
# $2: path to chrooted /lib64
# WARNING: does not copy binary itself!
install_chroot_program()
{
  local prg cp_update filetype libdir libdir64
  prg="$1"
  libdir="$2"
  libdir64="$3"

  filetype=`file -ib "$prg"`

  case "$filetype" in
  application/x-executable*statically' 'linked*)  ;;
  application/x-shellscript*|text/x-shellscript*) ;;
  application/x-executable*|application/x-sharedlib*)
    # sometimes 'file' say the executable is a shared library
    # newer OSes also say just "executable" without distibguishind between static and dynamic
    ldd "$prg" | awk '/not found/ { print; exit 1; }
            /not a dynamic executable/ { exit 0; }
            NR=4 && $1 !~ /^\// && $2=="=>" && $4 ~ /\(0x[[:xdigit:]]*\)/ {
                if ($3 ~ /\/lib64\//) {
                dir = DIR64;
                } else {
                dir = DIR;
                }
                if (system("cp -v -p " CP_UPDATE " " $3 " " dir "/" $1)!=0)
                { exit 1; }
              }
          ' DIR="$libdir" DIR64="$libdir64" CP_UPDATE="$cp_update"
    ;;
  *)
    echo "WARNING: $prg is not a program (filetype $filetype). Skipping." >&2
    return 2
    ;;
  esac
}

# Description: install basic chroot
# Parameters: none
# If executed with already existing chrooted template, will update binaries and libraries. 
# Useful after OS updates.
install_chroot_base()
{
  local libdir libdir64

  CHROOT_PROGRAMS="bash cat cp grep head id less ln ls mkdir more mv pwd rm rmdir sash sh tail touch du vi true false scp groups"
  
  os=`uname -s`
  case "$os" in
    Linux)
      cp_update=-u
      libdir="$CHROOT_ROOT_D/lib"
      libdir64="$CHROOT_ROOT_D/lib64"
      ;;
    FreeBSD)
      cp_update=-n
      libdir="$CHROOT_ROOT_D/usr/lib"
      ;;
    *) die "unknown system $os" ;;
  esac
  copy="cp -v -p $cp_update"

  mkdir -m 755 -p "$CHROOT_ROOT_D"
  chown 0:0 "$CHROOT_ROOT_D"

  mkdir -m 755 -p "$CHROOT_ROOT_D/bin"
  mkdir -m 755 -p "$CHROOT_ROOT_D/dev"
  mkdir -m 755 -p "$CHROOT_ROOT_D/lib"
  [ -d /lib64 ] && mkdir -m 755 -p "$CHROOT_ROOT_D/lib64"
  mkdir -m 755 -p "$CHROOT_ROOT_D/usr"
  mkdir -m 755 -p "$CHROOT_ROOT_D/usr/lib"
  [ -d /usr/libexec ] && mkdir -m 755 -p "$CHROOT_ROOT_D/usr/libexec"
  [ -d /libexec ] && mkdir -m 755 -p "$CHROOT_ROOT_D/libexec"
  mkdir -m 755 -p "$CHROOT_ROOT_D/etc"
  mkdir -m 1777 -p "$CHROOT_ROOT_D/tmp"
  mkdir -m 755 -p "$CHROOT_ROOT_D/var"
  mkdir -m 1777 -p "$CHROOT_ROOT_D/var/tmp"

  case "$os" in
  Linux)
    mknod -m 666 "$CHROOT_ROOT_D/dev/null" c 1 3
    chown root:root "$CHROOT_ROOT_D/dev/null"
    ;;
  FreeBSD)
    mknod "$CHROOT_ROOT_D/dev/null" c 1 3 root:wheel
    chmod 666 "$CHROOT_ROOT_D/dev/null"
    ;;
  *) die "unknown system $os" ;;
  esac

  cd "$CHROOT_ROOT_D/bin"
  for f in $CHROOT_PROGRAMS; do
    path=`which "$f"`
    found=$?
    if test $found -eq 0 -a -x "$path"; then
      cp -v -p "$path" "$f" 
      install_chroot_program "$f" "$libdir" "$libdir64"
    fi
  done

  sftp_path="$(awk '/^Subsystem[[:space:]]sftp.*$/ {print $3}' < /etc/ssh/sshd_config)"
  sftp_dir="${sftp_path%/*}"

  if [ -z "$sftp_dir" ]; then
    echo "Warning: sftp-server binary not found" 
    echo "+ sftp connections will not be available for chrooted accounts" 
    warn "installation of sftp-server binary into chroot environment"
  else
    mkdir -p "$CHROOT_ROOT_D$sftp_dir" 
    cp -v -p "$sftp_path" "$CHROOT_ROOT_D$sftp_path" 
    install_chroot_program "$CHROOT_ROOT_D$sftp_path" "$libdir" "$libdir64"
  fi

  case "$os" in
  Linux)
    $copy /lib/ld-linux* /lib/libnss_*.so.2  "$CHROOT_ROOT_D/lib" 
    if [ -d /lib64 ]; then
      $copy /lib64/ld-linux* /lib64/libnss_*.so.2  "$CHROOT_ROOT_D/lib64" 
    fi
    ;;
  FreeBSD)
    $copy /usr/libexec/ld-elf.so.* "$CHROOT_ROOT_D/usr/libexec" 
    $copy /libexec/ld-elf.so.* "$CHROOT_ROOT_D/libexec" 
    mkdir -m 755 -p "$CHROOT_ROOT_D/usr/share"
    mkdir -m 755 -p "$CHROOT_ROOT_D/usr/share/misc"
    cp -v -p /etc/termcap "$CHROOT_ROOT_D/usr/share/misc/termcap" 
    ;;
  esac

  if test -r /etc/termcap; then
    cp -v -p /etc/termcap "$CHROOT_ROOT_D/etc/termcap" 
  fi
  touch "$CHROOT_ROOT_D/etc/passwd"
  touch "$CHROOT_ROOT_D/etc/group"

  # noschg flag does not allow hardlinking security
  if test "$os" = "FreeBSD"; then
    chflags -R noschg "$CHROOT_ROOT_D"
  fi
}

bindir=$CHROOT_ROOT_D/bin
libdir=$CHROOT_ROOT_D/lib
libdir64=$CHROOT_ROOT_D/lib64

case "$1" in
  --install|--update)
    install_chroot_base && \
      echo "SUCCESS. Do not forget to run '$0 --apply domains...' to apply changes in chroot template to domains."
  ;;

  --add)
    prg=$2

    if [ ! -e $prg ] ; then
      echo "Error: file $prg is not found" >&2
      exit 1
    else
      install_chroot_program $prg $libdir $libdir64 && \
        cp -v -p -u $prg $bindir && \
        echo "SUCCESS. Do not forget to run '$0 --apply domains...' to apply changes in chroot template to domains."
    fi
  ;;

  --devices)
    add_devices="random urandom tty ptmx zero"
    MAKEDEV -d $CHROOT_ROOT_D/dev -D /dev -x -v $add_devices
    chmod 666 $CHROOT_ROOT_D/dev/*
    echo "SUCCESS. Do not forget to run '$0 --apply domains...' to apply changes in chroot template to domains."
  ;;

  --apply|--remove)
    list="$2"

    if [ -z "$list" ] ; then
      echo "ERROR: comma-separated list of domain names or word 'all' is required." >&2
      usage
      exit 1
    elif [ "$list" = "all" ] ; then
      list=$(all_domains)
    else
      list=${list//,/ }
    fi

    if [ "$1" == "--apply" ] ; then
        action="create"
        action_name="Updating"
    else
        action="remove"
        action_name="Removing"
    fi

    for domain in $list ; do
      # skip if domain does not exist or has no chrooted shell enabled
      domains_shell=$(q "SELECT s.shell FROM domains d, hosting h, sys_users s WHERE s.id = h.sys_user_id AND h.dom_id = d.id AND d.name = '$domain'")

      if [ -z "$domains_shell" -o "$domains_shell" != "$PRODUCT_ROOT_D/bin/chrootsh" ] ; then
        echo "WARNING: domain $domain does not exist or has no chrooted shell enabled. Skipping." >&2
        continue
      fi
      
      echo -n "$action_name chrooted environment on $domain: "
      $PRODUCT_ROOT_D/admin/sbin/chrootmng --$action \
        --source=$CHROOT_ROOT_D --target=$HTTPD_VHOSTS_D/$domain && echo "done"
    done
  ;;

  --help|"")
    usage
    exit 0
  ;;

  *)
    echo "Unknown option: $1" >&2
    exit 1
  ;;
esac
